
// Agent.fs - Aivika

// Copyright (c) 2009, 2010, 2011  David Sorokin. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//   * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
// 
//   * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials
//     provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
// THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
// OF THE POSSIBILITY OF SUCH DAMAGE.

//
// Author: David Sorokin <david.sorokin@gmail.com>
//

#nowarn "40"

namespace Maritegra.Aivika

open System
open System.Collections.Generic

open SD

type StateMode =
    | InitialMode
    | TransientMode
    | ProcessingMode

type InvalidAgentException (agent: Agent, message: string) =
    inherit Exception (message)
    member x.Agent = agent

and [<AbstractClass>] Agent (e: Env) as x = 

    let startState = 
        { new State (x, None) with
            override st.Activate = x.Activate 
            override st.Deactivate = x.Deactivate
        }
        
    let finalState =
        { new State (x, None) with
            override st.Activate = dynamics.Zero ()
            override st.Deactivate = dynamics {
                raise <| InvalidAgentException (x, "The agent cannot leave the final state")
            }
        }

    let dictLock = new obj ()
    let dictState = ref Map.empty
    let dictMode = ref Map.empty
    let dictFin = ref Map.empty
    
    let finalize (r: Run) =
        let k = r.Id
        if not (Map.containsKey k !dictFin) then
            lock dictLock (fun () ->
                dictFin := Map.add k () !dictFin)
            r.Disposed.Add (fun () ->
                lock dictLock (fun () ->
                    dictState := Map.remove k !dictState
                    dictMode := Map.remove k !dictMode
                    dictFin := Map.remove k !dictFin))

    let property defaultValue map =
        Dynamics (fun ps ->
            let r = ps.Run
            let k = r.Id
            if Map.containsKey k !map then
                let ri = Map.find k !map in !ri
            else
                defaultValue)
                    
    let setProperty value map = 
        Dynamics (fun ps ->
            let r = ps.Run
            let k = r.Id
            if Map.containsKey k !map then
                let ri = Map.find k !map
                ri := value
            else
                lock dictLock (fun () ->
                    map := Map.add k (ref value) !map)
                finalize r)

    let state = property startState dictState
    let setState value = setProperty value dictState

    let mode = property InitialMode dictMode
    let setMode value = setProperty value dictMode

    member x.Env = e
    
    member internal x.Mode = mode
    member internal x.SetMode value = setMode value
    
    member x.StartState = startState
    member x.FinalState = finalState
    member x.State = state
    
    member x.SetState (st: State) =
        assert (x = st.Agent)
        Dynamics (fun ps ->
            let m = mode <|- ps
            match m with
            | InitialMode ->
                raise <| InvalidAgentException 
                    (x, "The agent is in an initial mode. Use the SetInitialState method " +
                        "instead of SetState inside of the Activate property")
            | TransientMode ->
                raise <| InvalidAgentException 
                    (x, "The agent is in a transient mode. Use the SetInitialState method " +
                        "instead of SetState inside of the Activate property")
            | ProcessingMode ->
                let st0 = state <|- ps
                (st0.GoTo st) <|- ps)
        
    member x.SetInitialState (st: State) =
        assert (x = st.Agent)
        Dynamics (fun ps ->
            let m = mode <|- ps
            match m with
            | InitialMode ->
                let st0 = state <|- ps
                (st0.GoTo st) <|- ps
            | TransientMode -> ()
            | ProcessingMode ->
                raise <| InvalidAgentException 
                    (x, "The agent is in a processing mode. Use the SetState method " +
                        "instead of SetInitialState outside of the Activate property"))

    member internal x.SetStateImpl (st: State) = setState st
    
    abstract Activate: Dynamics<unit>
    abstract Deactivate: Dynamics<unit>

    default x.Activate = dynamics.Zero ()
    default x.Deactivate = dynamics.Zero ()
    
    member x.Start (t: float) = 
        Dynamics.once <| 
            e.Dispatch (eta t, 
                Dynamics (fun ps ->
                    x.Activate <|- ps
                    (setMode ProcessingMode) <|- ps))
    
    member x.StartD (t: Dynamics<float>) = 
        Dynamics.once <| 
            e.Dispatch (t, 
                Dynamics (fun ps ->
                    x.Activate <|- ps
                    (setMode ProcessingMode) <|- ps))
        
    member x.SetTimeout (t, action) = startState.SetTimeout (t, action)
    member x.SetTimeoutD (t, action) = startState.SetTimeoutD (t, action)
        
    member x.SetTimer (t, action) = startState.SetTimer (t, action)
    member x.SetTimerD (t, action) = startState.SetTimerD (t, action)
        
and [<AbstractClass>] State internal (agent: Agent, parent: State option) as x = 

    let findPath (target: State) =
    
        assert (agent = target.Agent)
    
        let rec fullPath (st: State) acc =
            match st.Parent with
            | Some st' -> fullPath st' (st :: acc)
            | None -> st :: acc
            
        let path1 = fullPath x []
        let path2 = fullPath target []
        
        let rec partitionPath path1 path2 = 
            match path1, path2 with
            | (h1 :: t1), [h2] when h1 = h2 ->
                List.rev path1, path2
            | (h1 :: t1), (h2 :: t2) when h1 = h2 -> 
                partitionPath t1 t2
            | _ -> 
                List.rev path1, path2
                
        partitionPath path1 path2

    let dictLock = new obj ()
    let dictVersion = ref Map.empty
    let dictFin = ref Map.empty
    
    let finalize (r: Run) =
        let k = r.Id
        if not (Map.containsKey k !dictFin) then
            lock dictLock (fun () ->
                dictFin := Map.add k () !dictFin)
            r.Disposed.Add (fun () ->
                lock dictLock (fun () ->
                    dictVersion := Map.remove k !dictVersion
                    dictFin := Map.remove k !dictFin))

    let property defaultValue map =
        Dynamics (fun ps ->
            let r = ps.Run
            let k = r.Id
            if Map.containsKey k !map then
                let ri = Map.find k !map in !ri
            else
                defaultValue)
                
    let incVersion (r: Run) =
        let k = r.Id
        if Map.containsKey k !dictVersion then
            let ri = Map.find k !dictVersion
            incr ri
        else
            lock dictLock (fun () ->
                dictVersion := Map.add k (ref 1) !dictVersion)
            finalize r

    let version = property 0 dictVersion

    let goto (path1: State list) (path2: State list) (target: State) =
        Dynamics (fun ps ->
        
            let r = ps.Run

            (agent.SetMode (TransientMode)) <|- ps
            
            for st in path1 do
            
                (agent.SetStateImpl (st)) <|- ps
                st.Deactivate <|- ps

                // it makes all timeout and timer handlers obsolete                    
                st.IncVersion (r)
            
            for st in path2 do
            
                if st = target then
                    (agent.SetMode (InitialMode)) <|- ps
                    
                (agent.SetStateImpl (st)) <|- ps
                st.Activate <|- ps
                
                if st = target then
                    (agent.SetMode (ProcessingMode)) <|- ps)

    let setTimeout (t: Dynamics<float>) action =
        Dynamics (fun ps ->
            let v = version <|- ps
            let e = agent.Env
            let m = Dynamics (fun ps ->
                let v' = version <|- ps
                if v = v' then
                    action <|- ps)
            (e.Dispatch (t + time, m)) <|- ps)

    let setTimer (t: Dynamics<float>) action =
        Dynamics (fun ps ->
            let v = version <|- ps
            let e = agent.Env
            let rec m = Dynamics (fun ps ->
                let v' = version <|- ps
                if v = v' then
                    (e.Dispatch (t + time, m)) <|- ps
                    action <|- ps)
            (e.Dispatch (t + time, m)) <|- ps)

    new (parent: State) = State (parent.Agent, Some parent)
    
    new (agent: Agent) = State (agent.StartState)
    
    member x.Agent = agent
    
    member x.Parent = parent
    
    member private x.IncVersion (r: Run) = incVersion r
    
    member internal x.GoTo (target: State) =
        let path1, path2 = findPath target
        in goto path1 path2 target
    
    abstract Activate: Dynamics<unit>
    abstract Deactivate: Dynamics<unit>

    default x.Activate = dynamics.Zero ()
    default x.Deactivate = dynamics.Zero ()
       
    member x.SetTimeout (t, action) = setTimeout (eta t) action
    member x.SetTimeoutD (t, action) = setTimeout t action
    
    member x.SetTimer (t, action) = setTimer (eta t) action
    member x.SetTimerD (t, action) = setTimer t action
