
// Dynamics.fsi - Aivika

// Copyright (c) 2009, 2010, 2011  David Sorokin. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//   * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
// 
//   * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials
//     provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
// THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
// OF THE POSSIBILITY OF SUCH DAMAGE.

//
// Author: David Sorokin <david.sorokin@gmail.com>
//

namespace Maritegra.Aivika

open System

type internal Time = float

type internal Iteration = int

type internal Phase = int

type Method = Euler | RungeKutta2 | RungeKutta4

type Randomness = SimpleRnd | StrongRnd

type Specs = 
    { StartTime: float; StopTime: float; DT: float; 
      Method: Method; Randomness: Randomness }

[<RequireQualifiedAccess>]
[<CompilationRepresentation (CompilationRepresentationFlags.ModuleSuffix)>]
module internal Specs =

    val iterations: Specs -> Iteration
    val phases: Specs -> Phase
    val time: Specs -> Iteration -> Phase -> Time
    val inline neighborhood: Specs -> Time -> Time -> bool

type internal Run =
    new: unit -> Run
    member Id: int64
    member Disposed: IEvent<unit>
    member IsDisposed: bool
    interface IDisposable

type internal Parameters =
    { Run: Run; Specs: Specs; Time: float;
      Iteration: Iteration; Phase: Phase }

type internal DynamicsFun<'a> = Parameters -> 'a

[<Sealed>] 
type Dynamics<'a> =
    internal new: DynamicsFun<'a> -> Dynamics<'a>
    member internal Fun: DynamicsFun<'a>

[<AutoOpen>]
module internal DynamicsOp =
    val inline (<|-): Dynamics<'a> -> DynamicsFun<'a>

[<Sealed>] 
type Dynamics<'a> with

    static member (~+): Dynamics<float> -> Dynamics<float>
    static member (~-): Dynamics<float> -> Dynamics<float>

    static member (+): Dynamics<float> * Dynamics<float> -> Dynamics<float>
    static member (+): Dynamics<float> * float -> Dynamics<float>
    static member (+): float * Dynamics<float> -> Dynamics<float>

    static member (-): Dynamics<float> * Dynamics<float> -> Dynamics<float>
    static member (-): Dynamics<float> * float -> Dynamics<float>
    static member (-): float * Dynamics<float> -> Dynamics<float>

    static member (*): Dynamics<float> * Dynamics<float> -> Dynamics<float>
    static member (*): Dynamics<float> * float -> Dynamics<float>
    static member (*): float * Dynamics<float> -> Dynamics<float>

    static member (/): Dynamics<float> * Dynamics<float> -> Dynamics<float>
    static member (/): Dynamics<float> * float -> Dynamics<float>
    static member (/): float * Dynamics<float> -> Dynamics<float>

    static member Abs: Dynamics<float> -> Dynamics<float>
    static member Acos: Dynamics<float> -> Dynamics<float>
    static member Asin: Dynamics<float> -> Dynamics<float>
    static member Atan: Dynamics<float> -> Dynamics<float>
    static member Atan2: Dynamics<float> * Dynamics<float> -> Dynamics<float>
    static member Ceiling: Dynamics<float> -> Dynamics<float>
    static member Exp: Dynamics<float> -> Dynamics<float>
    static member Floor: Dynamics<float> -> Dynamics<float>
    static member Truncate: Dynamics<float> -> Dynamics<float>
    static member Round: Dynamics<float> -> Dynamics<float>
    static member Log: Dynamics<float> -> Dynamics<float>
    static member Log10: Dynamics<float> -> Dynamics<float>
    static member Sqrt: Dynamics<float> -> Dynamics<float>
    static member Cos: Dynamics<float> -> Dynamics<float>
    static member Cosh: Dynamics<float> -> Dynamics<float>
    static member Sin: Dynamics<float> -> Dynamics<float>
    static member Sinh: Dynamics<float> -> Dynamics<float>
    static member Tan: Dynamics<float> -> Dynamics<float>
    static member Tanh: Dynamics<float> -> Dynamics<float>
    static member Pow: Dynamics<float> * Dynamics<float> -> Dynamics<float>
    static member Pow: Dynamics<float> * float -> Dynamics<float>

[<Sealed>]
type DynamicsBuilder =

    member Return: 'a -> Dynamics<'a>
    member ReturnFrom: Dynamics<'a> -> Dynamics<'a>
    member Bind: Dynamics<'a> * ('a -> Dynamics<'b>) -> Dynamics<'b>
    member Delay: (unit -> Dynamics<'b>) -> Dynamics<'b>
    member Zero: unit -> Dynamics<unit>
    member Combine: Dynamics<unit> * Dynamics<'a> -> Dynamics<'a>
    member For: seq<'a> * ('a -> Dynamics<unit>) -> Dynamics<unit>
    member While: (unit -> bool) * Dynamics<unit> -> Dynamics<unit>
    member Using: 'a * ('a -> Dynamics<'b>) -> Dynamics<'b> when 'a :> IDisposable
    member TryFinally: Dynamics<'a> * (unit -> unit) -> Dynamics<'a>
    member TryWith: Dynamics<'a> * (exn -> Dynamics<'a>) -> Dynamics<'a>

[<AutoOpen>]
module DynamicsWorkflow =     
    val dynamics: DynamicsBuilder

[<RequireQualifiedAccess>]
[<CompilationRepresentation (CompilationRepresentationFlags.ModuleSuffix)>]
module Dynamics =

    [<CompiledName ("Run")>]
    val run: Dynamics<'a> -> Specs -> seq<'a>
    
    [<CompiledName ("RunLast")>]
    val runLast: Dynamics<'a> -> Specs -> 'a
    
    [<CompiledName ("RunWhile")>]
    val runWhile: Dynamics<bool> -> Dynamics<'a> -> Specs -> seq<'a>
    
    [<CompiledName ("Eta")>]
    val eta: 'a -> Dynamics<'a>
    
    [<CompiledName ("Mu")>]
    val mu: Dynamics<Dynamics<'a>> -> Dynamics<'a>
    
    [<CompiledName ("Map")>]
    val map: ('a -> 'b) -> Dynamics<'a> -> Dynamics<'b>
    
    [<CompiledName ("Ap")>]
    val ap: Dynamics<'a -> 'b> -> Dynamics<'a> -> Dynamics<'b>
    
    [<CompiledName ("Lift2")>]
    val lift2: ('a -> 'b -> 'c) -> Dynamics<'a> -> Dynamics<'b> -> Dynamics<'c>
    
    [<CompiledName ("Lift3")>]
    val lift3: ('a -> 'b -> 'c -> 'd) -> Dynamics<'a> -> Dynamics<'b> -> Dynamics<'c> -> Dynamics<'d>

    [<CompiledName ("Lift4")>]
    val lift4: ('a -> 'b -> 'c -> 'd -> 'e) -> Dynamics<'a> -> Dynamics<'b> -> Dynamics<'c> -> Dynamics<'d> -> Dynamics<'e>
    
    [<CompiledName ("Lift5")>]
    val lift5: ('a -> 'b -> 'c -> 'd -> 'e -> 'f) -> Dynamics<'a> -> Dynamics<'b> -> Dynamics<'c> -> Dynamics<'d> -> Dynamics<'e> -> Dynamics<'f>
    
    [<CompiledName ("Generate")>]
    val generate: 'a -> Dynamics<'a -> 'b> -> Dynamics<'b>      // for side-effect
    
    [<CompiledName ("Zip")>]
    val zip: Dynamics<'a> -> Dynamics<'b> -> Dynamics<'a * 'b>
    
    [<CompiledName ("Zip3")>]
    val zip3: Dynamics<'a> -> Dynamics<'b> -> Dynamics<'c> -> Dynamics<'a * 'b * 'c>
    
    [<CompiledName ("Unzip")>]
    val unzip: Dynamics<'a * 'b> -> Dynamics<'a> * Dynamics<'b>
    
    [<CompiledName ("Unzip3")>]
    val unzip3: Dynamics<'a * 'b * 'c> -> Dynamics<'a> * Dynamics<'b> * Dynamics<'c>
    
    [<CompiledName ("UnzipFst")>]
    val unzipFst: Dynamics<'a * 'b> -> Dynamics<'a>
    
    [<CompiledName ("UnzipSnd")>]
    val unzipSnd: Dynamics<'a * 'b> -> Dynamics<'b>
    
    [<CompiledName ("Delay")>]
    val delay: Dynamics<'a> -> Dynamics<unit -> 'a>
    
    [<CompiledName ("OfList")>]
    val ofList: Dynamics<'b> list -> Dynamics<'b list>
    
    [<CompiledName ("OfArray")>]
    val ofArray: Dynamics<'b> [] -> Dynamics<'b []>
        
    [<CompiledName ("SeqList")>]
    val seqList: #seq<Dynamics<'b>> -> Dynamics<'b list>

    [<CompiledName ("SeqArray")>]
    val seqArray: #seq<Dynamics<'b>> -> Dynamics<'b []>
    
    [<CompiledName ("Init")>]
    val init: Dynamics<'a> -> Dynamics<'a>
    
    [<CompiledName ("Discrete")>]
    val discrete: Dynamics<'a> -> Dynamics<'a>

    [<CompiledName ("Interpolate")>]
    val interpolate: Dynamics<float> -> Dynamics<float>
    
    [<CompiledName ("Memo")>]
    val memo: (Dynamics<'a> -> Dynamics<'a>) -> Dynamics<'a> -> Dynamics<'a>
    
    [<CompiledName ("Memo0")>]
    val memo0: (Dynamics<'a> -> Dynamics<'a>) -> Dynamics<'a> -> Dynamics<'a>

    [<CompiledName ("MemoGenerator")>]
    val memoGenerator: (Specs -> 'a) -> Dynamics<'a -> 'b> -> Dynamics<'b>
    
    [<CompiledName ("Once")>]
    val once: Dynamics<unit> -> Dynamics<unit>
