
// SystemDynamics.fs - Aivika

// Copyright (c) 2009, 2010  David Sorokin. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//   * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
// 
//   * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials
//     provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
// THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
// OF THE POSSIBILITY OF SUCH DAMAGE.

//
// Author: David Sorokin <david.sorokin@gmail.com>
//

#nowarn "40"

namespace Maritegra.Aivika

open System
open System.Security.Cryptography

module SD = 

    [<CompiledName ("Eta")>]
    let eta a = Dynamics (fun ps -> a)

    [<CompiledName ("StartTime")>]
    let starttime = Dynamics (fun ps -> ps.Specs.StartTime)
    
    [<CompiledName ("StopTime")>]
    let stoptime = Dynamics (fun ps -> ps.Specs.StopTime)
    
    [<CompiledName ("DT")>]
    let dt = Dynamics (fun ps -> ps.Specs.DT)
    
    [<CompiledName ("Time")>]
    let time = Dynamics (fun ps -> ps.Time)
    
    [<CompiledName ("Init")>]
    let init f = Dynamics.init f
   
    [<CompiledName ("Discrete")>]
    let discrete f = Dynamics.discrete f
    
    type IntegImpl =
    
        static member Euler (y: Lazy<Dynamics<_>>, f: Lazy<Dynamics<_>>, i: Dynamics<_>, ps: Parameters) = 
        
            let s  = ps.Specs
            let n  = ps.Iteration
            let ph = ps.Phase
            
            match ph with
                | 0 ->
                    if n = 0 then
                        let vi = i <|- ps
                        in vi
                    else
                        let ty = Specs.time s (n-1) 0
                        let t1 = ty
                        
                        let psy = { ps with Time = ty; Iteration = n-1; Phase = 0 }
                        let ps1 = psy
                        
                        let vy = y.Value <|- psy
                        let k1 = f.Value <|- ps1
                        
                        in vy + s.DT * k1
                | _ -> 
                    failwithf "integ (Euler): incorrect phase = %i" ph 
    
        static member RK2 (y: Lazy<Dynamics<_>>, f: Lazy<Dynamics<_>>, i: Dynamics<_>, ps: Parameters) = 
        
            let s  = ps.Specs
            let n  = ps.Iteration
            let ph = ps.Phase
            
            match ph with
                | 0 ->
                    if n = 0 then
                        let vi = i <|- ps
                        in vi
                    else
                        let ty = Specs.time s (n-1) 0
                        let t1 = ty
                        let t2 = Specs.time s (n-1) 1
                        
                        let psy = { ps with Time = ty; Iteration = n-1; Phase = 0 }
                        let ps1 = psy
                        let ps2 = { ps with Time = t2; Iteration = n-1; Phase = 1 }
                        
                        let vy = y.Value <|- psy
                        let k1 = f.Value <|- ps1
                        let k2 = f.Value <|- ps2
                        
                        in vy + s.DT/2.0 * (k1 + k2)
                | 1 ->
                    let ty = Specs.time s n 0
                    let t1 = ty
                    
                    let psy = { ps with Time = ty; Iteration = n; Phase = 0 }
                    let ps1 = psy
                    
                    let vy = y.Value <|- psy
                    let k1 = f.Value <|- ps1
                    
                    in vy + s.DT * k1
                | _ -> 
                    failwithf "integ (RK2): incorrect phase = %i" ph 
    
        static member RK4 (y: Lazy<Dynamics<_>>, f: Lazy<Dynamics<_>>, i: Dynamics<_>, ps: Parameters) = 
        
            let s  = ps.Specs
            let n  = ps.Iteration
            let ph = ps.Phase
        
            match ph with
                | 0 -> 
                    if n = 0 then
                        let vi = i <|- ps
                        in vi
                    else
                        let ty = Specs.time s (n-1) 0
                        let t1 = ty
                        let t2 = Specs.time s (n-1) 1
                        let t3 = Specs.time s (n-1) 2
                        let t4 = Specs.time s (n-1) 3
                        
                        let psy = { ps with Time = ty; Iteration = n-1; Phase = 0 }
                        let ps1 = psy
                        let ps2 = { ps with Time = t2; Iteration = n-1; Phase = 1 }
                        let ps3 = { ps with Time = t3; Iteration = n-1; Phase = 2 }
                        let ps4 = { ps with Time = t4; Iteration = n-1; Phase = 3 }
                        
                        let vy = y.Value <|- psy
                        let k1 = f.Value <|- ps1
                        let k2 = f.Value <|- ps2
                        let k3 = f.Value <|- ps3
                        let k4 = f.Value <|- ps4
                        
                        in vy + s.DT/6.0 * (k1 + 2.0*k2 + 2.0*k3 + k4)
                | 1 ->   
                    let ty = Specs.time s n 0
                    let t1 = ty
                    
                    let psy = { ps with Time = ty; Iteration = n; Phase = 0 }
                    let ps1 = psy
                    
                    let vy = y.Value <|- psy
                    let k1 = f.Value <|- ps1
                    
                    in vy + s.DT/2.0 * k1
                | 2 ->
                    let ty = Specs.time s n 0
                    let t2 = Specs.time s n 1
                    
                    let psy = { ps with Time = ty; Iteration = n; Phase = 0 }
                    let ps2 = { ps with Time = t2; Iteration = n; Phase = 1 }
                    
                    let vy = y.Value <|- psy
                    let k2 = f.Value <|- ps2
                    
                    in vy + s.DT/2.0 * k2
                | 3 ->
                    let ty = Specs.time s n 0
                    let t3 = Specs.time s n 2
                    
                    let psy = { ps with Time = ty; Iteration = n; Phase = 0 }
                    let ps3 = { ps with Time = t3; Iteration = n; Phase = 2 }
                    
                    let vy = y.Value <|- psy
                    let k3 = f.Value <|- ps3
                    in vy + s.DT * k3
                | _ -> 
                    failwithf "integ (RK4): incorrect phase = %i" ph 
    
    [<CompiledName ("Integ")>]
    let integ (f: Lazy<Dynamics<float>>) (i: Dynamics<float>) =
    
        let rec y = lazy (Dynamics.memo Dynamics.interpolate z)
        and z = Dynamics (fun ps -> 
            match ps.Specs.Method with
                | Euler -> IntegImpl.Euler (y, f, i, ps)
                | RungeKutta2 -> IntegImpl.RK2 (y, f, i, ps)
                | RungeKutta4 -> IntegImpl.RK4 (y, f, i, ps))
        in y.Value
        
    [<CompiledName ("Smooth")>]
    let smooth (x: Dynamics<float>) (t: Lazy<Dynamics<float>>) =
        let rec y = integ (lazy ((x - y) / t.Value)) x in y
        
    [<CompiledName ("SmoothI")>]
    let smoothI (x: Lazy<Dynamics<float>>) (t: Lazy<Dynamics<float>>) (i: Dynamics<float>) =
        let rec y = integ (lazy ((x.Value - y) / t.Value)) i in y
        
    [<CompiledName ("Smooth3")>]
    let smooth3 (x: Dynamics<float>) (t: Lazy<Dynamics<float>>) =
        
        let rec y = integ (lazy ((s1 - y) / (t.Value / 3.0))) x
        and s1 = integ (lazy ((s0 - s1) / (t.Value / 3.0))) x
        and s0 = integ (lazy ((x - s0) / (t.Value / 3.0))) x
        in y
        
    [<CompiledName ("Smooth3I")>]
    let smooth3I (x: Lazy<Dynamics<float>>) (t: Lazy<Dynamics<float>>) (i: Dynamics<float>) =
        
        let rec y = integ (lazy ((s1 - y) / (t.Value / 3.0))) i
        and s1 = integ (lazy ((s0 - s1) / (t.Value / 3.0))) i
        and s0 = integ (lazy ((x.Value - s0) / (t.Value / 3.0))) i
        in y
        
    [<CompiledName ("SmoothN")>]
    let smoothN (x: Dynamics<float>) (t: Lazy<Dynamics<float>>) n =
    
        let rec s = [| 
            for k = 0 to n-1 do
                if k = 0 then
                    yield integ (lazy ((x - s.[k]) / (t.Value / (float n)))) x
                else
                    yield integ (lazy ((s.[k-1] - s.[k]) / (t.Value / (float n)))) x |]
        in s.[n-1]
        
    [<CompiledName ("SmoothNI")>]
    let smoothNI (x: Lazy<Dynamics<float>>) (t: Lazy<Dynamics<float>>) n (i: Dynamics<float>) =
    
        let rec s = [| 
            for k = 0 to n-1 do
                if k = 0 then
                    yield integ (lazy ((x.Value - s.[k]) / (t.Value / (float n)))) i
                else
                    yield integ (lazy ((s.[k-1] - s.[k]) / (t.Value / (float n)))) i |]
        in s.[n-1]
        
    [<CompiledName ("Delay1")>]
    let delay1 (x: Dynamics<float>) (t: Dynamics<float>) =
        let rec y = integ (lazy (x - y)) (x*t) / t in y
        
    [<CompiledName ("Delay1I")>]
    let delay1I (x: Lazy<Dynamics<float>>) (t: Dynamics<float>) (i: Dynamics<float>) =
        let rec y = integ (lazy (x.Value - y)) (i*t) / t in y
        
    [<CompiledName ("Delay3")>]
    let delay3 (x: Dynamics<float>) (t: Dynamics<float>) =
        
        let rec y = integ (lazy (s1 - y)) (x*(t/3.0)) / (t/3.0)
        and s1 = integ (lazy (s0 - s1)) (x*(t/3.0)) / (t/3.0)
        and s0 = integ (lazy (x - s0)) (x*(t/3.0)) / (t/3.0)
        in y
        
    [<CompiledName ("Delay3I")>]
    let delay3I (x: Lazy<Dynamics<float>>) (t: Dynamics<float>) (i: Dynamics<float>) =
        
        let rec y = integ (lazy (s1 - y)) (i*(t/3.0)) / (t/3.0)
        and s1 = integ (lazy (s0 - s1)) (i*(t/3.0)) / (t/3.0)
        and s0 = integ (lazy (x.Value - s0)) (i*(t/3.0)) / (t/3.0)
        in y
        
    [<CompiledName ("DelayN")>]
    let delayN (x: Dynamics<float>) (t: Dynamics<float>) n =
    
        let rec s = [| 
            for k = 0 to n-1 do
                if k = 0 then
                    yield integ (lazy (x - s.[k])) (x*(t/(float n))) / (t/(float n))
                else
                    yield integ (lazy (s.[k-1] - s.[k])) (x*(t/(float n))) / (t/(float n)) |]
        in s.[n-1]
        
    [<CompiledName ("DelayNI")>]
    let delayNI (x: Lazy<Dynamics<float>>) (t: Dynamics<float>) n (i: Dynamics<float>) =
    
        let rec s = [| 
            for k = 0 to n-1 do
                if k = 0 then
                    yield integ (lazy (x.Value - s.[k])) (i*(t/(float n))) / (t/(float n))
                else
                    yield integ (lazy (s.[k-1] - s.[k])) (i*(t/(float n))) / (t/(float n)) |]
        in s.[n-1]
        
    [<CompiledName ("Forecast")>]
    let forecast (x: Dynamics<float>) (at: Dynamics<float>) (hz: Dynamics<float>) =
        x * (1.0 + (x / smooth x (lazy at) - 1.0) / at * hz)
        
    [<CompiledName ("Trend")>]
    let trend (x: Dynamics<float>) (at: Dynamics<float>) (i: Dynamics<float>) =
        (x / smoothI (lazy x) (lazy at) (x / (1.0 + i * at)) - 1.0) / at
            
    type Reservoir (init: Dynamics<float>) =
            
        let mutable inflow = eta 0.0
        let mutable outflow = eta 0.0
        
        let diff = Dynamics (fun ps ->
            (inflow <|- ps) - (outflow <|- ps))
        
        let value = integ (lazy diff) init
    
        new (init: double) = Reservoir (dynamics {return init})
        
        member x.Inflow with get() = inflow and set (v) = inflow <- v
        member x.Outflow with get() = outflow and set (v) = outflow <- v
        member x.Value = value

    type Table (ps: (float * float) []) =
    
        let last = (Array.length ps) - 1
    
        let rec find left right x =
        
            if left > right then nan
            else
            
                let index = (left + 1 + right) / 2
                let x1 = fst ps.[index]
                
                if x1 <= x then 
                
                    if index < right then find index right x
                    elif right = last then snd ps.[right]
                    else
                    
                        let x2 = fst ps.[index + 1]
                        let y1 = snd ps.[index]
                        let y2 = snd ps.[index + 1]
                        in y1 + (y2 - y1) * (x - x1) / (x2 - x1)
                    
                else
                
                    if left < index then find left (index - 1) x
                    elif left = 0 then snd ps.[left]
                    else nan
                    
        let rec findStepwise left right x =
        
            if left > right then nan
            else
            
                let index = (left + 1 + right) / 2
                let x1 = fst ps.[index]
                
                if x1 <= x then
                
                    if index < right then findStepwise index right x
                    elif right = last then snd ps.[right]
                    else  snd ps.[index]
                    
                else
                
                    if left < index then findStepwise left (index - 1) x
                    elif left = 0 then snd ps.[left]
                    else nan

        member tbl.LookupF (x: float) : float = find 0 last x
        member tbl.LookupStepwiseF (x: float) = findStepwise 0 last x
        
        member tbl.Lookup (x: Dynamics<float>) = 
            Dynamics (fun ps -> find 0 last (x <|- ps))

        member tbl.LookupStepwise (x: Dynamics<float>) = 
            Dynamics (fun ps -> findStepwise 0 last (x <|- ps))
    
    [<CompiledName ("Table")>]
    let table ps = Table ps

    [<CompiledName ("Lookup")>]
    let lookup (x: Dynamics<float>) (t: Table)  = t.Lookup x
    
    [<CompiledName ("LookupF")>]
    let lookupF (x: float) (t: Table) = t.LookupF x

    [<CompiledName ("LookupStepwise")>]
    let lookupStepwise (x: Dynamics<float>) (t: Table) = t.LookupStepwise x
    
    [<CompiledName ("LookupStepwiseF")>]
    let lookupStepwiseF (x: float) (t: Table) = t.LookupStepwiseF x

    type Generator = unit -> float

    let simpleGenerator seed =
        match seed with
            | None ->
                let r = Random ()
                in fun () -> r.NextDouble ()
            | Some s ->
                let r = Random (s)
                in fun () -> r.NextDouble ()
                            
    let strongGenerator () =
        let r = new RNGCryptoServiceProvider()
        let d = Array.create 8 0uy
        in fun () ->
            
                r.GetBytes (d)
                
                let i1 = uint64 (d.[0] &&& 3uy)
                let i1 = (i1 <<< 8) + uint64 d.[1]
                let i1 = (i1 <<< 8) + uint64 d.[2]
                let i1 = (i1 <<< 8) + uint64 d.[3]
                
                let i2 = uint64 (d.[4] &&& 7uy)
                let i2 = (i2 <<< 8) + uint64 d.[5]
                let i2 = (i2 <<< 8) + uint64 d.[6]
                let i2 = (i2 <<< 8) + uint64 d.[7]
                
                in (float ((i1 <<< 27) + i2)) / ((float) (1L <<< 53))

    [<CompiledName ("Randomizer")>]
    let randomizer seed (tr: Generator -> Generator) (m: Dynamics<Generator -> 'a>) =
        fun s ->
            match seed with
                | Some x as y when x <> 0 -> 
                    simpleGenerator y
                | _ -> 
                    match s.Randomness with
                        | StrongRnd -> strongGenerator ()
                        | _ -> simpleGenerator None
        |> (fun f -> f >> tr)
        |> (fun f -> Dynamics.memoGenerator f m)
        
    [<CompiledName ("Random")>]
    let random (a: Dynamics<float>) (b: Dynamics<float>) =
        Dynamics (fun ps ->
            let a' = a <|- ps
            let b' = b <|- ps
            in fun g -> a' + (b' - a') * g ())
        |> randomizer None id

    [<CompiledName ("RandomS")>]
    let randomS s (a: Dynamics<float>) (b: Dynamics<float>) = 
        Dynamics (fun ps ->
            let a' = a <|- ps
            let b' = b <|- ps
            in fun g -> a' + (b' - a') * g ())
        |> randomizer (Some s) id

    let normalGenerator g =
    
        let next = ref 0.0
        let flag = ref false
        
        in fun () ->
        
            if !flag then
            
                flag := false
                !next
                
            else
            
                let mutable xi1 = 0.0
                let mutable xi2 = 0.0
                let mutable psi = 0.0
                
                while (psi >= 1.0) || (psi = 0.0) do

                    xi1 <- 2.0 * g () - 1.0
                    xi2 <- 2.0 * g () - 1.0
                    psi <- xi1 * xi1 + xi2 * xi2

                psi <- sqrt (- 2.0 * (log psi) / psi)
                
                flag := true
                next := xi2 * psi
                
                xi1 * psi
                
    [<CompiledName ("Normal")>]
    let normal (mu: Dynamics<float>) (sigma: Dynamics<float>) = 
        Dynamics (fun ps ->
            let mu' = mu <|- ps
            let sigma' = sigma <|- ps
            in fun g -> mu' + sigma' * g ())
        |> randomizer None normalGenerator

    [<CompiledName ("NormalS")>]
    let normalS s (mu: Dynamics<float>) (sigma: Dynamics<float>) = 
        Dynamics (fun ps ->
            let mu' = mu <|- ps
            let sigma' = sigma <|- ps
            in fun g -> mu' + sigma' * g ())
        |> randomizer (Some s) normalGenerator

    let binomialGenerator prob trials g =
        let mutable sum = 0
        for i = 1 to trials do
            if g () <= prob then sum <- sum + 1
        sum

    [<CompiledName ("Binomial")>]
    let binomial (prob: Dynamics<float>) (trials: Dynamics<int>) = 
        Dynamics (fun ps ->
            let prob' = prob <|- ps
            let trials' = trials <|- ps
            in binomialGenerator prob' trials')
        |> randomizer None id

    [<CompiledName ("BinomialS")>]
    let binomialS s (prob: Dynamics<float>) (trials: Dynamics<int>) = 
        Dynamics (fun ps ->
            let prob' = prob <|- ps
            let trials' = trials <|- ps
            in binomialGenerator prob' trials')
        |> randomizer (Some s) id

    let poissonGenerator mu g =
    
        let mutable v = 0
        let mutable prob = g ()
        let mutable prod = exp (- mu)
        
        while (prob > prod) do

            v <- v + 1
            prob <- prob - prod
            prod <- prod * mu / (float v)
            
        v
        
    [<CompiledName ("Poisson")>]
    let poisson (mu: Dynamics<float>) = 
        Dynamics (fun ps ->
            let mu' = mu <|- ps
            in poissonGenerator mu')
        |> randomizer None id
        
    [<CompiledName ("PoissonS")>]
    let poissonS s (mu: Dynamics<float>) = 
        Dynamics (fun ps ->
            let mu' = mu <|- ps
            in poissonGenerator mu')
        |> randomizer (Some s) id
        
    [<CompiledName ("ExpRnd")>]
    let exprnd (mu: Dynamics<float>) = 
        Dynamics (fun ps ->
            let mu' = mu <|- ps
            in fun g -> - log (g ()) * mu')
        |> randomizer None id
        
    [<CompiledName ("ExpRndS")>]
    let exprndS s (mu: Dynamics<float>) = 
        Dynamics (fun ps ->
            let mu' = mu <|- ps
            in fun g -> - log (g ()) * mu')
        |> randomizer (Some s) id

    [<CompiledName ("Delay")>]
    let delay (x: Dynamics<'a>) (d: Dynamics<float>) = 
    
        Dynamics (fun ps ->
        
            let t = ps.Time
            let s = ps.Specs
            let n = ps.Iteration
        
            let t' = t - (d <|- ps)
            let t' = t' - s.StartTime
            let n' = int (floor (t' / s.DT))
            
            if n' < 0 then
                x <|- { ps with Time = s.StartTime; Iteration = 0; Phase = 0 }
            elif n' < n then
                x <|- { ps with Time = t'; Iteration = n'; Phase = -1 }
            elif n' > n then
                failwith "delay: cannot return the future data"
            else
                failwith "delay: cannot return the current data")
                
        |> discrete

    [<CompiledName ("DelayI")>]
    let delayI (x: Lazy<Dynamics<'a>>) (d: Dynamics<float>) (i: Dynamics<'a>) = 
    
        Dynamics (fun ps ->
        
            let t = ps.Time
            let s = ps.Specs
            let n = ps.Iteration
        
            let t' = t - (d <|- ps)
            let t' = t' - s.StartTime
            let n' = int (floor (t' / s.DT))
            
            if n' < 0 then
                i <|- { ps with Time = s.StartTime; Iteration = 0; Phase = 0 }
            elif n' < n then
                x.Value <|- { ps with Time = t'; Iteration = n'; Phase = -1 }
            elif n' > n then
                failwith "delayI: cannot return the future data"
            else
                failwith "delayI: cannot return the current data")
                
        |> Dynamics.memo0 discrete

    [<CompiledName ("Step")>]
    let step (h: Dynamics<float>) (st: Dynamics<float>) = 
        Dynamics (fun ps ->
            let s = ps.Specs
            let t = ps.Time
            let st' = st <|- ps
            let t'  = t + s.DT / 2.0
            if st' < t' then
                h <|- ps
            else
                0.0) 
        |> discrete

    [<CompiledName ("Pulse")>]
    let pulse (st: Dynamics<float>) (w: Dynamics<float>) = 
        Dynamics (fun ps ->
            let s = ps.Specs
            let t = ps.Time
            let st' = st <|- ps
            let t'  = t + s.DT / 2.0
            if st' < t' then
                let w' = w <|- ps
                if t' < st' + w' then 1.0 else 0.0
            else
                0.0) 
        |> discrete
        
    [<CompiledName ("PulseP")>]
    let pulseP (st: Dynamics<float>) (w: Dynamics<float>) (p: Dynamics<float>) =
        Dynamics (fun ps ->
            let s = ps.Specs
            let t = ps.Time
            let p'  = p <|- ps
            let st' = st <|- ps
            let y'  = if p' > 0.0 && t > st' then round ((t - st') / p') * p' else 0.0
            let st' = st' + y'
            let t'  = t + s.DT / 2.0
            if st' < t' then
                let w' = w <|- ps
                if t' < st' + w' then 1.0 else 0.0
            else
                0.0) 
        |> discrete
        
    [<CompiledName ("Ramp")>]
    let ramp (slope: Dynamics<float>) (st: Dynamics<float>) (e: Dynamics<float>) =
        Dynamics (fun ps ->
            let s = ps.Specs
            let t = ps.Time
            let st' = st <|- ps
            let t1' = t + s.DT / 2.0
            if st' < t1' then
                let slope' = slope <|- ps
                let e'  = e <|- ps
                let t2' = t - s.DT / 2.0
                if t2' < e' then
                    slope' * (t - st')
                else
                    slope' * (e' - st')
            else
                0.0)
        |> discrete

    [<CompiledName ("SinWave")>]
    let sinWave a p = a * sin (2.0 * Math.PI * time / p)
    
    [<CompiledName ("CosWave")>]
    let cosWave a p = a * cos (2.0 * Math.PI * time / p)
    
    [<CompiledName ("MaxD")>]
    let maxD x1 x2 =
        Dynamics (fun ps ->
            let x1' = x1 <|- ps
            let x2' = x2 <|- ps
            in max x1' x2')
    
    [<CompiledName ("MinD")>]
    let minD x1 x2 =
        Dynamics (fun ps ->
            let x1' = x1 <|- ps
            let x2' = x2 <|- ps
            in min x1' x2')

    [<CompiledName ("NonNegative")>]
    let nonnegative (m: Dynamics<float>) = 
        Dynamics (fun ps ->
            let x = m <|- ps
            in max 0.0 x)
