﻿
(* An agent-based version of the classical Bass Diffusion model. 
   You can find a similar model in the documentation of 
   AnyLogic [http://www.xjtek.com] *)

#nowarn "40"

open System

open Maritegra.Aivika
open Maritegra.Aivika.SD

let n = 500     // the number of agents

let advertisingEffectiveness = 0.011
let contactRate = 100.0
let adoptionFraction = 0.015

let specs =
    { StartTime = 0.0; StopTime = 8.0; DT = 0.1;
      Method = RungeKutta4; Randomness = SimpleRnd }

let random = 
    let rnd = new Random ()
    in fun a b -> rnd.Next (a, b)
      
let randomTrue = 
    let rnd = new Random ()
    in fun p -> rnd.NextDouble () <= p

let expovariate = 
    let rnd = new Random ()
    in fun lambda ->
        - log (rnd.NextDouble ()) / lambda 
        
let e = Env ()

let potentialAdopters = Var (e, 0)
let adopters = Var (e, 0)

type Person () as agent =
    
    inherit Agent (e)
    
    static let mutable agents = ResizeArray<Person> ()
    
    let rec potentialAdopter =
        { new State (agent) with
            member x.Activate = dynamics {
                do! Var.incValue 1 potentialAdopters
                
                // create a timeout that will hold while the state is active 
                do! x.SetTimeout (expovariate advertisingEffectiveness,
                        agent.SetState (adopter))
            }
            
            member x.Deactivate = dynamics {
                do! Var.decValue 1 potentialAdopters
            }
        }
    and adopter =
        { new State (agent) with
            member x.Activate = dynamics {
                do! Var.incValue 1 adopters
                
                // create a timer that will hold while the state is active
                do! x.SetTimerD (dynamics { 
                                    return expovariate contactRate 
                                 },
                                 dynamics {
                                    do! agents.[random 0 (agents.Count - 1)].Buy
                                 })
            }
            
            member x.Deactivate = dynamics {
                do! Var.decValue 1 adopters
            }
        }
        
    member private x.Buy = dynamics {
        let! st = x.State
        if st = potentialAdopter then
            if randomTrue adoptionFraction then
                do! x.SetState (adopter)
    }
    
    override x.Activate = dynamics {
        agents.Add (x)
        do! x.SetInitialState (potentialAdopter)
    }

[<EntryPoint>]
let main args =

    let starter = 
        dynamics {
            for i = 1 to n do
                let agent = Person ()
                do! agent.StartD (starttime)
        } |> Dynamics.once
    
    let model = 
        dynamics {
            do! starter     // N.B. it is called only once
                
            let! t  = time
            let! a1 = potentialAdopters.Value
            let! a2 = adopters.Value
            
            return (t, a1, a2)
        }
    
    let results = Dynamics.run model specs
    
    printfn "Time,Potential Adopters,Adopters"
    
    for (t, a1, a2) in results do
        printfn "%f,%i,%i" t a1 a2
    
    0