﻿
(* This is a rewritten version of model MachRep1 based on
   the event-oriented paradigm of DES. *)

(* It corresponds to model MachRep1 described in document 
   Introduction to Discrete-Event Simulation and the SimPy Language
   [http://heather.cs.ucdavis.edu/~matloff/156/PLN/DESimIntro.pdf]. 
   SimPy is available on [http://simpy.sourceforge.net/].
   
   The model description is as follows.

   Two machines, which sometimes break down.
   Up time is exponentially distributed with mean 1.0, and repair time is
   exponentially distributed with mean 0.5. There are two repairpersons,
   so the two machines can be repaired simultaneously if they are down
   at the same time.

   Output is long-run proportion of up time. Should get value of about
   0.66. *)

#nowarn "40"

open System

open Maritegra.Aivika
open Maritegra.Aivika.SD

let specs = {

    StartTime=0.0; StopTime=10000.0; DT=0.05; 
    Method=RungeKutta4; Randomness=SimpleRnd
}

let upRate = 1.0 / 1.0          // reciprocal of mean up time
let repairRate = 1.0 / 0.5      // reciprocal of mean repair time

let mutable totalUpTime = 0.0   // total up time for all machines

let expovariate = 
    let rnd = new Random ()
    in fun lambda ->
        - log (rnd.NextDouble ()) / lambda 
        
let machine id (e: Env) = 

    let rec broken startUpTime =
        dynamics {
        
            // the machine is broken
            let! finishUpTime = time
            totalUpTime <- totalUpTime 
                + (finishUpTime - startUpTime)
            let repairTime = expovariate repairRate
            
            // register a new event
            do! e.Dispatch (eta (finishUpTime + repairTime), 
                    repaired)
        }
    and repaired =
        dynamics {
        
            // the machine is repaired
            let! startUpTime = time
            let upTime = expovariate upRate
            
            // register a new event
            do! e.Dispatch (eta (startUpTime + upTime),
                    broken startUpTime)    // closure
        }
    and started =
        e.Dispatch (starttime, repaired)
        
    in Dynamics.once started
    
let e = Env ()    
    
let machine1 = machine 1 e
let machine2 = machine 2 e

let starter = 
    dynamics {
        do! machine1
        do! machine2
    } |> Dynamics.once

let system = dynamics {

    do! starter     // start the machines
    do! e.Run       // run the event queue
    
    return! (totalUpTime / (2.0 * stoptime))
}

[<EntryPoint>]
let main args =

    let result = Dynamics.runLast system specs
    
    printfn "Long-run proportion of up time = %f (must be about 0.66)" result
    
    0
