﻿
(* This is an advanced version of model MachRep1. 
   The simulations can be launched in parallel. 
   What is more important is that the represented 
   DES submodel can be integrated into the hybrid model. *)

(* It corresponds to model MachRep1 described in document 
   Introduction to Discrete-Event Simulation and the SimPy Language
   [http://heather.cs.ucdavis.edu/~matloff/156/PLN/DESimIntro.pdf]. 
   SimPy is available on [http://simpy.sourceforge.net/].
   
   The model description is as follows.

   Two machines, which sometimes break down.
   Up time is exponentially distributed with mean 1.0, and repair time is
   exponentially distributed with mean 0.5. There are two repairpersons,
   so the two machines can be repaired simultaneously if they are down
   at the same time.

   Output is long-run proportion of up time. Should get value of about
   0.66. *)

open System

open Maritegra.Aivika
open Maritegra.Aivika.SD

let specs = {

    StartTime=0.0; StopTime=10000.0; DT=1.0; 
    Method=RungeKutta4; Randomness=SimpleRnd
}

let upRate = 1.0 / 1.0          // reciprocal of mean up time
let repairRate = 1.0 / 0.5      // reciprocal of mean repair time

let expovariate = 
    let rnd = new Random ()
    in fun lambda ->
        - log (rnd.NextDouble ()) / lambda 

let env = Env ()

// total up time for all machines
let totalUpTime = Var<float> (env, 0.0)

type Machine (e, id) =

    inherit Process (e)
    
    override x.Activate = dynamicscont {
        while true do
            let! startUpTime = DynamicsCont.lift time
            let upTime = expovariate upRate
            do! x.Hold (upTime)
            let! finishUpTime = DynamicsCont.lift time
            do! DynamicsCont.lift (totalUpTime |> 
                    Var.incValue (finishUpTime - startUpTime))
            let repairTime = expovariate repairRate
            do! x.Hold (repairTime)
    }

let machines = 
    [| for i = 1 to 2 do 
        yield Machine (env, i) |]

let starter = 
    dynamics {
        for m in machines do
            do! m.StartD (starttime)    // launch the machine
    } |> Dynamics.once

let system = dynamics {
    do! starter     // launch the machines            
    do! env.Run     // launch the environment
    return! (totalUpTime.Value / (2.0 * stoptime))
}

[<EntryPoint>]
let main args =

    let result = Dynamics.runLast system specs
    printfn "Long-run proportion of up time = %f (must be about 0.66)" result
    
    0
