﻿
(* It corresponds to model MachRep2 described in document 
   Introduction to Discrete-Event Simulation and the SimPy Language
   [http://heather.cs.ucdavis.edu/~matloff/156/PLN/DESimIntro.pdf]. 
   SimPy is available on [http://simpy.sourceforge.net/].
   
   The model description is as follows.
   
   Two machines, but sometimes break down. Up time is exponentially 
   distributed with mean 1.0, and repair time is exponentially distributed 
   with mean 0.5. In this example, there is only one repairperson, so 
   the two machines cannot be repaired simultaneously if they are down 
   at the same time.

   In addition to finding the long-run proportion of up time as in
   model MachRep1, let’s also find the long-run proportion of the time 
   that a given machine does not have immediate access to the repairperson 
   when the machine breaks down. Output values should be about 0.6 and 0.67. *)

open System

open Maritegra.Aivika
open Maritegra.Aivika.SD

let specs = {

    StartTime=0.0; StopTime=10000.0; DT=1.0; 
    Method=RungeKutta4; Randomness=SimpleRnd
}
    
let upRate = 1.0 / 1.0          // breakdown rate
let repairRate = 1.0 / 0.5      // repair rate

// number of times the machines have broken down
let mutable nRep = 0            

// number of breakdonws in which the machine 
// started repair service right away
let mutable nImmedRep = 0       

// total up time for all machines
let mutable totalUpTime = 0.0   

let expovariate = 
    let rnd = new Random ()
    in fun lambda ->
        - log (rnd.NextDouble ()) / lambda 

type Machine (e, repairPerson: Resource, id) =

    inherit Process (e)
    
    override x.Activate = dynamicscont {
        while true do
            let! startUpTime = DynamicsCont.lift time
            do! x.Hold (expovariate upRate)
            let! finishUpTime = DynamicsCont.lift time
            totalUpTime <- totalUpTime 
                + (finishUpTime - startUpTime)
            nRep <- nRep + 1
            //
            let! n = repairPerson.N
            if n = 1 then
                nImmedRep <- nImmedRep + 1
            do! repairPerson.Request
            do! x.Hold (expovariate repairRate)
            do! repairPerson.Release
    }

let env = Env ()

let repairPerson = Resource (env, 1)

let machines = 
    [| for i = 1 to 2 do 
        yield Machine (env, repairPerson, i) |]

let starter = 
    dynamics {
        for m in machines do
            do! m.StartD (starttime)    // launch the machine
    } |> Dynamics.once

let system = dynamics {

    do! starter         // launch the machines
    do! env.Run         // launch the environment
    
    let m1 = (totalUpTime / (2.0 * stoptime))
    let m2 = (float nImmedRep) / (float nRep)
    
    return! Dynamics.zip m1 (eta m2)
}

[<EntryPoint>]
let main args =

    let result = Dynamics.runLast system specs
    
    printfn "Long-run proportion of up time = %f (must be about 0.6)" (fst result)
    printfn "Long-run proportion of the time when \
        immediate access to the repairperson = %f (must be about 0.67)" (snd result)
    
    0
