﻿
(* It corresponds to model MachRep3 described in document 
   Introduction to Discrete-Event Simulation and the SimPy Language
   [http://heather.cs.ucdavis.edu/~matloff/156/PLN/DESimIntro.pdf]. 
   SimPy is available on [http://simpy.sourceforge.net/].
   
   The model description is as follows.

   Variation of models MachRep1, MachRep2. Two machines, but
   sometimes break down. Up time is exponentially distributed with mean
   1.0, and repair time is exponentially distributed with mean 0.5. In
   this example, there is only one repairperson, and she is not summoned
   until both machines are down. We find the proportion of up time. It
   should come out to about 0.45. *)

#nowarn "40"

open System

open Maritegra.Aivika
open Maritegra.Aivika.SD

let specs = {

    StartTime=0.0; StopTime=10000.0; DT=1.0; 
    Method=RungeKutta4; Randomness=SimpleRnd
}
    
let upRate = 1.0 / 1.0          // breakdown rate
let repairRate = 1.0 / 0.5      // repair rate

// number of machines currently up
let mutable nUp = 0

// total up time for all machines
let mutable totalUpTime = 0.0

let expovariate = 
    let rnd = new Random ()
    in fun lambda ->
        - log (rnd.NextDouble ()) / lambda 

type Machine (e, repairPerson: Resource, m2: Lazy<Machine>, id) =

    inherit Process (e)
    
    do nUp <- nUp + 1
    
    override x.Activate = dynamicscont {
        while true do
            let! startUpTime = DynamicsCont.lift time
            do! x.Hold (expovariate upRate)
            let! finishUpTime = DynamicsCont.lift time
            totalUpTime <- totalUpTime 
                + (finishUpTime - startUpTime)
            //
            nUp <- nUp - 1
            if nUp = 1 then
                do! x.Passivate
            else
                let! n = repairPerson.N
                if n = 1 then
                    do! m2.Value.Reactivate
            do! repairPerson.Request
            do! x.Hold (expovariate repairRate)
            nUp <- nUp + 1
            do! repairPerson.Release
    }

let env = Env ()

let repairPerson = Resource (env, 1)

let rec machines = [| 
    for i = 1 to 2 do 
        yield Machine (env, repairPerson, 
            lazy (machines.[1 - (i - 1)]), i) |]

let starter = 
    dynamics {
        for m in machines do
            do! m.StartD (starttime)    // launch the machine
    } |> Dynamics.once

let system = dynamics {

    do! starter         // launch the machines
    do! env.Run         // launch the environment
    
    return! (totalUpTime / (2.0 * stoptime))
}

[<EntryPoint>]
let main args =

    let result = Dynamics.runLast system specs
    printfn "The proportion of up time = %f (must be about 0.45)" result
    
    0
