﻿
(* This model is described in the Russian book stated below.
  
   Книга "Объектно-ориентированное моделирование на C++", 2006. Автор Илья Труб.
   Модель описана в главе 7 "Последовательное обслуживание с блокировками
   и ограниченным буфером: производственная поточная линия". 
   
   Можно обратить внимание на то, как вычисляется среднее значение
   через интеграл. То есть, модель фактически является гибридной. 
   Используются методы DES и системной динамики вместе. *)

#nowarn "40"

open System
open System.Collections.Generic

open Maritegra.Aivika
open Maritegra.Aivika.SD

module Workplace =

    let specs = {
    
        StartTime=0.0; StopTime=300.0; DT=1.00; 
        Method=Euler; Randomness=SimpleRnd
    }
    
    let inline lift m = DynamicsCont.lift m
    
    let mean m = integ (lazy m) (eta 0.0) / (time - starttime)
    
    let expovariate = 
        let rnd = new Random ()
        in fun lambda ->
            - log (rnd.NextDouble ()) / lambda 
            
    let sizeWk1 = 4
    let sizeWk2 = 2
        
    let lambdaWk1 = 1.0 / 0.25
    let lambdaWk2 = 1.0 / 0.5
    let lambdaLine = 1.0 / 0.4
            
    let env = Env ()
    
    let totalNum = Var (env, 0)
    let acceptNum = Var (env, 0)
    let rejectNum = Var (env, 0)
    
    let queueWk1 = Var (env, 0.0)
    let queueWk2 = Var (env, 0.0)
    
    let queueWk1Mean = mean queueWk1.Value
    let queueWk2Mean = mean queueWk2.Value
    
    let processSum = Var (env, 0.0)
    let processTimeMean = processSum.Value / Dynamics.map float acceptNum.Value
    
    let blockWk1 = Var (env, 0.0)
    let blockWk1Mean = mean blockWk1.Value
    
    let loadWk1 = Var (env, 0.0)
    let loadWk2 = Var (env, 0.0)
    
    let loadWk1Mean = mean loadWk1.Value
    let loadWk2Mean = mean loadWk2.Value
    
    type Workplace1 (e, wk2: Lazy<Workplace2>) =
    
        inherit Process (e)
        
        let queue = ResizeArray<_> (sizeWk1 + 1)
        
        member x.CanLoad () = queue.Count < sizeWk1 + 1
        
        member x.Load () = dynamicscont {
            let! t = time |> lift
            queue.Add (t)
            do! queueWk1.SetValue (float queue.Count - 1.0) |> lift
            if queue.Count = 1 then
                do! x.Reactivate
        }
        
        member x.Pass (t) = dynamicscont {
            if wk2.Value.CanLoad () then
                do! wk2.Value.Load (t)
            else
                do! blockWk1.SetValue (1.0) |> lift
                do! x.Passivate
                do! blockWk1.SetValue (0.0) |> lift
                do! x.Pass (t)
        }
        
        override x.Activate = dynamicscont {
            while true do
                if queue.Count = 0 then
                    do! x.Passivate
                else
                    do! loadWk1.SetValue (1.0) |> lift
                    do! x.Hold (expovariate lambdaWk1)
                    do! loadWk1.SetValue (0.0) |> lift
                    //
                    do! x.Pass (queue.[0])
                    queue.RemoveAt (0)
                    do! queueWk1.SetValue (max 0.0 (float queue.Count - 1.0)) |> lift
        }
        
    and Workplace2 (e, wk1: Lazy<Workplace1>) =
    
        inherit Process (e)
        
        let queue = ResizeArray<_> (sizeWk2 + 1)
        
        member x.CanLoad () = queue.Count < sizeWk2 + 1
        
        member x.Load (t1) = dynamicscont {
            let! t2 = time |> lift
            queue.Add ((t1, t2))
            do! queueWk2.SetValue (float queue.Count - 1.0) |> lift
            if queue.Count = 1 then 
                do! x.Reactivate
        }
        
        member x.Pass (t1, t2) = dynamicscont {
            let! t = time |> lift
            do! processSum |> Var.incValue (t - t1) |> lift
        }
        
        override x.Activate = dynamicscont {
            while true do
                if queue.Count = 0 then
                    do! x.Passivate
                else
                    do! loadWk2.SetValue (1.0) |> lift
                    do! x.Hold (expovariate lambdaWk2)
                    do! loadWk2.SetValue (0.0) |> lift
                    //
                    do! x.Pass queue.[0]
                    queue.RemoveAt (0)
                    do! queueWk2.SetValue (max 0.0 (float queue.Count - 1.0)) |> lift
                    //
                    do! wk1.Value.Reactivate
        }
        
    and Line (e, wk1: Lazy<Workplace1>, wk2: Lazy<Workplace2>) =
    
        inherit Process (e)
        
        override x.Activate = dynamicscont {
            while true do
                do! x.Hold (expovariate lambdaLine)
                do! totalNum |> Var.incValue 1 |> lift
                if wk1.Value.CanLoad () then
                    do! acceptNum |> Var.incValue 1 |> lift
                    do! wk1.Value.Load ()
                else
                    do! rejectNum |> Var.incValue 1 |> lift
        }
        
    let rec wk1 = Workplace1 (env, lazy wk2)
    and wk2 = Workplace2 (env, lazy wk1)
    and line = Line (env, lazy wk1, lazy wk2)
    
    let starter =
        dynamics {
            do! wk1.StartD (starttime)
            do! wk2.StartD (starttime)
            do! line.StartD (starttime)
        } |> Dynamics.once
    
    let system = dynamics {
    
        do! starter
        do! env.Run
        
        let! x1  = totalNum.Value
        let! x2  = acceptNum.Value
        let! x3  = rejectNum.Value
        let! x4  = blockWk1Mean
        let  x5  = ()
        let! x6  = loadWk1Mean
        let! x7  = loadWk2Mean
        let! x8  = processTimeMean
        let! x9  = queueWk1Mean
        let! x10 = queueWk2Mean
        let  x11 = ()
        
        return (x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11)
    }

    [<EntryPoint>]    
    let main args =
    
        let (x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11) = Dynamics.runLast system specs
    
        printfn "total number: %i" x1
        printfn "accepted number: %i" x2
        printfn "rejected number: %i" x3
        printfn "block rate in wk1: %f" x4
        printfn "block duration in wk1: ---"
        printfn "load rate in wk1: %f" x6
        printfn "load rate in wk2: %f" x7
        printfn "processing time: %f" x8
        printfn "queue size in wk1: %f" x9
        printfn "queue size in wk2: %f" x10
        
        0
